//
// bridge.java
//
// This application fixes a problem with Illustra. When connecting to
// Illustra over a socket connection, client read operations may block forever.
// This is not the case when running the client on the server. This
// bridge runs at port 7532 and routes all incoming traffic to port
// 7531 (Illustra). Similar for outgoing traffic.
//
// I.e.:
//
//  client   ---------> server:7532  -----------> server:7531
//           <--------               <-----------
//
//
// A production version is written as a daemon, called bridged.c
//
// Usage: java bridge
// 
// Author: Erwin Kalvelagen
// Date: July 29, 1997
//
// This version is not reliable, probably because sockets are not
// threadsafe. The c version however seems to work OK.

import java.net.*;
import java.io.*;


public class bridge {

  final static int ILLUSTRA_PORT = 7531;          // port Illustra is listening to
  final static int BRIDGE_PORT = 7532;            // port bridge is listening to
  final static String ILLUSTRA_HOST = "localhost";  // hostname of Illustra server

  final static boolean debug = true;  // if true logging output is printed to stdout

  final static boolean TCP_NODELAY = false;
  
  public static void main(String argv[]) {




    ServerSocket ss; // listens to outside world
    
    try {
      
      
      // first setup a server socket that listens to port 7532

      ss = new ServerSocket(BRIDGE_PORT);
      if (debug)
	System.out.println("Listening to port "+ss.getLocalPort());


      while (true) {

	// wait for incoming connections

	Socket s = ss.accept();
	if (debug) 
	  System.out.println("Incoming connection "+s);

	if (TCP_NODELAY)
	  s.setTcpNoDelay(true);

	// s.setSoTimeout(5000);



	// spawn a thread to handle it

	BridgeThread t = new BridgeThread(s, ILLUSTRA_HOST, ILLUSTRA_PORT);
	t.start();
	
      } // while
      
    } // try

    catch (Exception e) {
      System.err.println(e);
    }

  } // main

} //class bridge



class BridgeThread extends Thread {

  // this class implements a thread for a single remote client
  // it spawns two threads, one for each direction (incoming
  // traffic is forwarded to Illustra and outgoing traffic from
  // from Illustra is forwarded to the remote client).

  Socket outside;      // incoming connection from remote client
  Socket illustra;     // connection to illustra on local server
  String Host;         // "localhost"
  int Port;            // Illustra's port number
  


  public BridgeThread(Socket outside, String Host, int Port) {

    // Constructor

    this.outside = outside;
    this.Host = Host;
    this.Port = Port;

  }


  public void run() {


    try {
      
      // first setup a connection to Illustra on this host
      
      illustra = new Socket(Host, Port);
      if (bridge.debug)
	System.out.println("Connected to Illustra: "+illustra);

      
      // start threads
      // one for each direction

      MoverThread m1 = new MoverThread(outside.getInputStream(), 
				       outside.getInetAddress().toString(),
				       illustra.getOutputStream(), 
				       illustra.getInetAddress().toString());
      m1.start();
      MoverThread m2 = new MoverThread(illustra.getInputStream(), 
				       illustra.getInetAddress().toString(),
				       outside.getOutputStream(), 
				       outside.getInetAddress().toString());
      m2.start();

      // wait for them to end.....

      try {
	m1.join();
	m2.join();
      } 
      catch (InterruptedException e) {}

      // close sockets

      illustra.close();
      outside.close();

    } // try
    catch (Exception e) {
      System.err.println(e);
    }
  } // run

} // class BridgeThread





class MoverThread extends Thread {


  // this class implements a simple forwarder.
  // It reads bytes from 'is' and forwards them to 'os'

  InputStream is;
  OutputStream os;
  String name_is;
  String name_os;

  public MoverThread(InputStream is, String name_is, OutputStream os, String name_os) {
    this.is = is;
    this.os = os;
    this.name_is = name_is;
    this.name_os = name_os;
  }


  public void run() {
    
    byte[] buffer = new byte[8192];
    int n;
    
    
    try {
      
      while(true) {
	
	n = is.read(buffer); // will block until some bytes are read
	if (bridge.debug)
	  System.out.println("Read from "+name_is+" "+n+" bytes");
	if (n == -1) // eof?
	  break;     // then done
	
	if (n>0) {
	  System.out.println("Trying to write to "+name_os+" "+n+" bytes");
	  os.write(buffer,0,n);
	  os.flush();
	  if (bridge.debug)
	    System.out.println("Wrote to "+name_os+" "+n+" bytes");
	  
	}
      } // while
    } //try
    catch (Exception e) {
      System.out.println(e);
    }
  }// run
  
} // class MoverThread





